from unittest.mock import MagicMock

import pytest

from llamea import LLaMEA, Ollama_LLM, Solution


# Helper
class obj(object):
    def __init__(self, d):
        for k, v in d.items():
            if isinstance(k, (list, tuple)):
                setattr(self, k, [obj(x) if isinstance(x, dict) else x for x in v])
            else:
                setattr(self, k, obj(v) if isinstance(v, dict) else v)


def test_algorithm_generation():
    """Test the algorithm generation process."""

    def f(solution):
        return f"feedback {solution.name}", 1.0, "", {}

    optimizer = LLaMEA(
        f, llm=Ollama_LLM("model"), experiment_name="test generation", log=False
    )
    response = "# Description: Long Example Algorithm\n# Code:\n```python\nclass ExampleAlgorithm:\n    pass\n```"
    optimizer.llm.query = MagicMock(return_value=response)

    individual = optimizer.llm.sample_solution(
        session_messages=[{"role": "system", "content": "test prompt"}]
    )

    assert (
        individual.description == "Long Example Algorithm"
    ), f"Algorithm long name should be extracted correctly, is {individual.description}"
    assert (
        individual.name == "ExampleAlgorithm"
    ), "Algorithm name should be extracted correctly"
    assert (
        "class ExampleAlgorithm" in individual.code
    ), "Algorithm code should be extracted correctly"


def test_evolve_solution_with_diff():
    def f(sol, logger):
        return sol

    optimizer = LLaMEA(
        f,
        llm=Ollama_LLM("model"),
        experiment_name="diff",
        log=False,
        diff_mode=True,
        evaluate_population=False,
    )

    base = Solution(code="class MyAlgo:\n    pass\n", name="MyAlgo", description="d")
    optimizer.population = [base]
    diff_reply = """
```
<<<<<<< SEARCH
    pass
=======
    # Added loops
    for i in range(m):
        for k in range(n):
            for j in range(p):
                C[i, j] += A[i, k] * B[k, j]
>>>>>>> REPLACE
```"""
    optimizer.llm.query = MagicMock(return_value=diff_reply)
    evolved = optimizer.evolve_solution(base)
    assert "# Added loops" in evolved.code, evolved.code + " does not contain added code '# Added loops'"
    assert "pass" not in evolved.code, evolved.code + " contains code 'pass'"



# Uses Python Abstract Syntax Trees to extract graph characteristics
import argparse
import ast
import difflib
import json

import jellyfish
import jsonlines
import lizard
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import tqdm
from networkx.drawing.nx_pydot import graphviz_layout
from scipy.stats import entropy


def code_compare(code1, code2, printdiff=False):
    # Parse the Python code into ASTs
    # Use difflib to find differences
    diff = difflib.ndiff(code1.splitlines(), code2.splitlines())
    # Count the number of differing lines
    diffs = sum(1 for x in diff if x.startswith("- ") or x.startswith("+ "))
    # Calculate total lines for the ratio
    total_lines = max(len(code1.splitlines()), len(code2.splitlines()))
    similarity_ratio = (total_lines - diffs) / total_lines if total_lines else 1
    return 1 - similarity_ratio


def analyse_complexity(code):
    # Analyse the code complexity of the code
    i = lizard.analyze_file.analyze_source_code("algorithm.py", code)
    complexities = []
    token_counts = []
    parameter_counts = []
    for f in i.function_list:
        complexities.append(f.__dict__["cyclomatic_complexity"])
        token_counts.append(f.__dict__["token_count"])
        parameter_counts.append(len(f.__dict__["full_parameters"]))
    return {
        "mean_complexity": np.mean(complexities),
        "total_complexity": np.sum(complexities),
        "mean_token_count": np.mean(token_counts),
        "total_token_count": np.sum(token_counts),
        "mean_parameter_count": np.mean(parameter_counts),
        "total_parameter_count": np.sum(parameter_counts),
    }


# Parse Python AST and build a graph
class BuildAST(ast.NodeVisitor):
    def __init__(self):
        self.graph = nx.DiGraph()
        self.current_node = 0
        self.node_stack = []

    def generic_visit(self, node):
        node_id = self.current_node
        self.graph.add_node(node_id, label=type(node).__name__)

        if self.node_stack:
            parent_id = self.node_stack[-1]
            self.graph.add_edge(parent_id, node_id)

        self.node_stack.append(node_id)
        self.current_node += 1

        super().generic_visit(node)

        self.node_stack.pop()

    def build_graph(self, root):
        self.visit(root)
        return self.graph


def eigenvector_centrality_numpy(G, max_iter=500):
    try:
        return (nx.eigenvector_centrality_numpy(G, max_iter=500),)
    except Exception:
        return np.nan


# Function to extract graph characteristics
def analyze_graph(G):
    depths = dict(nx.single_source_shortest_path_length(G, min(G.nodes())))
    degrees = sorted((d for n, d in G.degree()), reverse=True)
    leaf_depths = [
        depth for node, depth in depths.items() if G.out_degree(node) == 0
    ]  # depth from root to leaves
    clustering_coefficients = list(nx.clustering(G).values())
    # Additional Features (not in paper)
    # Convert the directed graph to an undirected graph to avoid SCC problems
    undirected_G = G.to_undirected()
    if nx.is_connected(undirected_G):  # check if undirected graph is connected
        diameter = nx.diameter(undirected_G)
        radius = nx.radius(undirected_G)
        avg_shortest_path = nx.average_shortest_path_length(undirected_G)
        avg_eccentricity = np.mean(list(nx.eccentricity(undirected_G).values()))
    else:
        # Calculate diameter of the largest strongly connected component
        largest_cc = max(nx.connected_components(undirected_G), key=len)
        subgraph = G.subgraph(largest_cc)
        diameter = nx.diameter(subgraph)
        radius = nx.radius(subgraph)
        avg_shortest_path = nx.average_shortest_path_length(subgraph)
        avg_eccentricity = np.mean(list(nx.eccentricity(subgraph).values()))
    edge_density = (
        G.number_of_edges() / (G.number_of_nodes() * (G.number_of_nodes() - 1))
        if G.number_of_nodes() > 1
        else 0
    )

    return {
        # Number of Nodes and Edges
        "Nodes": G.number_of_nodes(),
        "Edges": G.number_of_edges(),
        # Degree Analysis
        # "Degrees": degrees,
        "Max Degree": max(degrees),
        "Min Degree": min(degrees),
        "Mean Degree": np.mean(degrees),
        "Degree Variance": np.var(degrees),
        # Transitivity
        "Transitivity": nx.transitivity(G),
        # Depth analysis
        # "Depths": leaf_depths,
        "Max Depth": max(leaf_depths),
        "Min Depth": min(leaf_depths),
        "Mean Depth": np.mean(leaf_depths),
        # Clustering Coefficients
        # "Clustering Coefficients": clustering_coefficients,
        "Max Clustering": max(clustering_coefficients),
        "Min Clustering": min(clustering_coefficients),
        "Mean Clustering": nx.average_clustering(G),
        "Clustering Variance": np.var(clustering_coefficients),
        # Entropy
        "Degree Entropy": entropy(degrees),
        "Depth Entropy": entropy(leaf_depths),
        # Additional features (not in paper)
        # "Betweenness Centrality": nx.betweenness_centrality(G),
        # "Eigenvector Centrality": eigenvector_centrality_numpy(G, max_iter=500),
        "Assortativity": nx.degree_assortativity_coefficient(G),
        "Average Eccentricity": avg_eccentricity,
        "Diameter": diameter,
        "Radius": radius,
        # "Pagerank": nx.pagerank(G, max_iter=500),
        "Edge Density": edge_density,
        "Average Shortest Path": avg_shortest_path,
    }


def visualize_graph(G):
    pos = graphviz_layout(G, prog="dot")
    labels = nx.get_node_attributes(G, "label")
    nx.draw(
        G,
        pos,
        labels=labels,
        with_labels=True,
        node_size=500,
        node_color="lightblue",
        font_size=8,
        font_weight="bold",
        arrows=True,
    )
    plt.savefig("graph1.pdf")


# Function to create graph out of AST
def process_file(path, visualize):
    with open(path, "r") as file:
        python_code = file.read()
    root = ast.parse(python_code)
    build = BuildAST()
    G = build.build_graph(root)
    stats = analyze_graph(G)

    complexity_stats = analyse_complexity(python_code)
    if visualize == True:  # visualize graph
        visualize_graph(G)
    return {**stats, **complexity_stats}


# Function to create graph out of AST
def process_code(python_code, visualize):
    root = ast.parse(python_code)
    build = BuildAST()
    G = build.build_graph(root)
    stats = analyze_graph(G)
    if visualize == True:  # visualize graph
        visualize_graph(G)

    complexity_stats = analyse_complexity(python_code)
    # print(stats)
    # print(complexity_stats)
    # print({**stats, **complexity_stats})
    # a = aaaa
    return {**stats, **complexity_stats}


def aggregate_stats(results):
    print("Aggregate Statistics:")
    print("Total Nodes:", sum(result["Nodes"] for result in results))
    print("Total Edges:", sum(result["Edges"] for result in results))
    print(
        "Average Transitivity:",
        sum(result["Transitivity"] for result in results) / len(results),
    )
    print("Max Depth:", max(result["Max Depth"] for result in results))
    print(
        "Average Degree Mean:", np.mean([result["Mean Degree"] for result in results])
    )
    print(
        "Average Clustering Coefficient:",
        np.mean([result["Mean Clustering"] for result in results]),
    )
    print(
        "Average Eccentricity:",
        np.mean([result["Average Eccentricity"] for result in results]),
    )
    print(
        "Average Edge Density:", np.mean([result["Edge Density"] for result in results])
    )


def print_results(stats, file):
    print("Statistics for file:", file)
    print("Number of nodes:", stats["Nodes"])
    print("Number of Edges:", stats["Edges"])
    print("Degrees:", stats["Degrees"])
    print("Maximum Degree:", stats["Max Degree"])
    print("Minimum Degree:", stats["Min Degree"])
    print("Mean Degree:", stats["Mean Degree"])
    print("Degree Variance:", stats["Degree Variance"])
    print("Transitivity:", stats["Transitivity"])
    print("Leaf Depths:", stats["Depths"])
    print("Max Depth:", stats["Max Depth"])
    print("Min Depth:", stats["Min Depth"])
    print("Mean Depth:", stats["Mean Depth"])
    print("Clustering Coefficients:", stats["Clustering Coefficients"])
    print("Max Clustering:", stats["Max Clustering"])
    print("Min Clustering:", stats["Min Clustering"])
    print("Mean Clustering:", stats["Mean Clustering"])
    print("Clustering Variance:", stats["Clustering Variance"])
    print("Degree Entropy:", stats["Degree Entropy"])
    print("Depth Entropy:", stats["Depth Entropy"])
    print("Betweenness Centrality:", stats["Betweenness Centrality"])
    print("Eigenvector Centrality:", stats["Eigenvector Centrality"])
    print("Assortativity:", stats["Assortativity"])
    print("Average Eccentricity:", stats["Average Eccentricity"])
    print("Diameter:", stats["Diameter"])
    print("Radius:", stats["Radius"])
    print("Pagerank:", stats["Pagerank"])
    print("Edge Density:", stats["Edge Density"])
    print("Average Shortest Path:", stats["Average Shortest Path"])
    print("")


def main(file_paths, visualize):
    results = []
    for file_path in file_paths:
        stats = process_file(file_path, visualize)
        results.append(stats)
        print_results(stats, file_path)
    # aggregate_stats(results)


if __name__ == "__main__":
    import os
    import sys

    budget = 100

    if True:  # switch to turn off some part of the code
        exp_dirs = [
            "runs/exp-09-06_122145-gpt-4o-2024-05-13-ES BP-HPO-long",
            "runs/exp-09-02_095606-gpt-4o-2024-05-13-ES BP-HPO-long",
            "runs/exp-08-30_141720-gpt-4o-2024-05-13-ES BP-HPO-long",
        ]

        colors = ["C0", "C0", "C1", "C1", "C2", "C2", "C3", "k"]
        linestyles = [
            "solid",
            "dotted",
            "solid",
            "dotted",
            "solid",
            "dotted",
            "solid",
            "solid",
        ]
        labels = ["LLaMEA-HPO"]

        best_ever_code = ""
        best_ever_name = ""
        best_ever_config = {}
        best_ever_fitness = -100
        results = []
        for i in range(len(exp_dirs)):
            convergence = np.ones(budget) * -100
            # convergence_default = np.zeros(budget)
            code_diff_ratios = np.zeros(budget)
            best_so_far = -np.inf
            best_so_far_default = 0
            previous_code = ""
            previous_config = {}
            previous_name = ""
            previous_gen = 0
            log_file = exp_dirs[i] + "/log.jsonl"
            if os.path.exists(log_file):
                with jsonlines.open(log_file) as reader:
                    for obj in reader.iter(type=dict, skip_invalid=True):
                        stats = []
                        gen = 0
                        fitness = None
                        code_diff = 0
                        code = ""
                        if "_solution" in obj.keys():
                            code = obj["_solution"]
                        if "_generation" in obj.keys():
                            gen = obj["_generation"]
                        if "_fitness" in obj.keys():
                            fitness = obj["_fitness"]
                            # print(fitness)
                        else:
                            fitness = None

                        if fitness > best_ever_fitness:
                            best_ever_fitness = fitness
                            best_ever_code = code
                            best_ever_config = obj["incumbent"]
                            best_ever_name = obj["_name"]

                        if fitness <= best_so_far:
                            code_diff = code_compare(previous_code, code, False)
                        else:
                            name = obj["_name"]
                            print(f"-- {gen} -- {previous_name} --> {name}")
                            code_diff = code_compare(previous_code, code, False)
                            best_so_far = fitness

                            previous_gen = gen
                            previous_code = code
                            previous_name = name
                            previous_config = obj["incumbent"]

                        code_diff_ratios[gen] = code_diff
                        convergence[gen] = fitness

                        try:
                            stats = process_code(code, False)
                        except Exception as e:
                            print(e)
                            continue

                        stats["code_diff"] = code_diff
                        stats["fitness"] = 0.0
                        stats["LLM"] = labels[0]
                        stats["exp_dir"] = exp_dirs[i]
                        stats["alg_id"] = gen
                        stats["parent_id"] = previous_gen
                        stats["fitness"] = fitness
                        results.append(stats)

        exp_dirs = ["run1", "run2", "run3"]
        convergence_lines = []
        for exp_dir in exp_dirs:
            conv_line = np.ones(budget * 200) * -np.inf
            best_so_far = -np.inf
            teller = 0
            gen = 0
            for k in range(5):
                with open(
                    "benchmarks/EoHresults/Prob1_OnlineBinPacking/"
                    + exp_dir
                    + f"/population_generation_{k}.json"
                ) as f:
                    pop = json.load(f)
                for ind in pop:
                    # if teller > budget:
                    #     break
                    if -1 * ind["objective"] > best_so_far:
                        best_so_far = -1 * ind["objective"]
                    conv_line[teller] = best_so_far
                    if k == 0:
                        teller += 1
                    else:
                        for x in range(5):  # EoH creates 5 offspring per individual
                            conv_line[teller] = best_so_far
                            teller += 1
                    code = ind["code"]
                    try:
                        stats = process_code(code, False)
                    except Exception:
                        continue

                    stats["code_diff"] = 0
                    stats["fitness"] = 0.0
                    stats["LLM"] = "EoH"
                    stats["exp_dir"] = exp_dir
                    stats["alg_id"] = gen
                    stats["parent_id"] = gen
                    stats["fitness"] = -1 * ind["objective"]
                    results.append(stats)
                    gen += 1

        convergence_lines.append(np.array(conv_line))

        import pandas as pd

        resultdf = pd.DataFrame.from_dict(results)
        print(resultdf)

        resultdf.to_csv("ast/graphstats_BP.csv", index=False)

        # TSP ------------------------------------------------------------

        exp_dirs = [
            "runs/exp-08-29_201655-gpt-4o-2024-05-13-ES TSP-HPO",
            "runs/exp-08-30_142330-gpt-4o-2024-05-13-ES TSP-HPO-deter",
            "runs/exp-09-02_105043-gpt-4o-2024-05-13-ES TSP-HPO-deter",
        ]

        colors = ["C0", "C0", "C1", "C1", "C2", "C2", "C3", "k"]
        linestyles = [
            "solid",
            "dotted",
            "solid",
            "dotted",
            "solid",
            "dotted",
            "solid",
            "solid",
        ]
        labels = ["LLaMEA-HPO"]

        best_ever_code = ""
        best_ever_name = ""
        best_ever_config = {}
        best_ever_fitness = -100
        results = []
        for i in range(len(exp_dirs)):
            convergence = np.ones(budget) * -100
            # convergence_default = np.zeros(budget)
            code_diff_ratios = np.zeros(budget)
            best_so_far = -np.inf
            best_so_far_default = 0
            previous_code = ""
            previous_config = {}
            previous_name = ""
            previous_gen = 0
            log_file = exp_dirs[i] + "/log.jsonl"
            if os.path.exists(log_file):
                with jsonlines.open(log_file) as reader:
                    for obj in reader.iter(type=dict, skip_invalid=True):
                        stats = []
                        gen = 0
                        fitness = None
                        code_diff = 0
                        code = ""
                        if "_solution" in obj.keys():
                            code = obj["_solution"]
                        if "_generation" in obj.keys():
                            gen = obj["_generation"]
                        if "_fitness" in obj.keys():
                            fitness = obj["_fitness"]
                            # print(fitness)
                        else:
                            fitness = None

                        if fitness > best_ever_fitness:
                            best_ever_fitness = fitness
                            best_ever_code = code
                            best_ever_config = obj["incumbent"]
                            best_ever_name = obj["_name"]

                        if fitness <= best_so_far:
                            code_diff = code_compare(previous_code, code, False)
                        else:
                            name = obj["_name"]
                            print(f"-- {gen} -- {previous_name} --> {name}")
                            code_diff = code_compare(previous_code, code, False)
                            best_so_far = fitness

                            previous_gen = gen
                            previous_code = code
                            previous_name = name
                            previous_config = obj["incumbent"]

                        code_diff_ratios[gen] = code_diff
                        convergence[gen] = fitness

                        try:
                            stats = process_code(code, False)
                        except Exception as e:
                            print(e)
                            continue

                        stats["code_diff"] = code_diff
                        stats["fitness"] = 0.0
                        stats["LLM"] = labels[0]
                        stats["exp_dir"] = exp_dirs[i]
                        stats["alg_id"] = gen
                        stats["parent_id"] = previous_gen
                        stats["fitness"] = fitness
                        results.append(stats)

        exp_dirs = ["run1", "run2", "run3"]
        convergence_lines = []
        for exp_dir in exp_dirs:
            conv_line = np.ones(budget * 200) * -np.inf
            best_so_far = -np.inf
            teller = 0
            gen = 0
            for k in range(5):
                with open(
                    "benchmarks/EoHresults/Prob2_TSP_GLS/"
                    + exp_dir
                    + f"/population_generation_{k}.json"
                ) as f:
                    pop = json.load(f)
                for ind in pop:
                    # if teller > budget:
                    #     break
                    if -1 * ind["objective"] > best_so_far:
                        best_so_far = -1 * ind["objective"]
                    conv_line[teller] = best_so_far
                    if k == 0:
                        teller += 1
                    else:
                        for x in range(5):  # EoH creates 5 offspring per individual
                            conv_line[teller] = best_so_far
                            teller += 1
                    code = ind["code"]
                    try:
                        stats = process_code(code, False)
                    except Exception:
                        continue

                    stats["code_diff"] = 0
                    stats["fitness"] = 0.0
                    stats["LLM"] = "EoH"
                    stats["exp_dir"] = exp_dir
                    stats["alg_id"] = gen
                    stats["parent_id"] = gen
                    stats["fitness"] = -1 * ind["objective"]
                    results.append(stats)
                    gen += 1

        convergence_lines.append(np.array(conv_line))

        import pandas as pd

        resultdf = pd.DataFrame.from_dict(results)
        print(resultdf)

        resultdf.to_csv("ast/graphstats_TSP.csv", index=False)

        # FIXED MUTATION RUNS ------------------------------

        base_dir = "ast/fixed_mutation_experiments/exp-prompt5-gpt-4o/"

        exp_dirs_2 = [
            "2/exp-11-01_150117-LLaMEA-gpt-4o-2",
            "2/exp-11-01_184846-LLaMEA-gpt-4o-2",
            "2/exp-11-01_211835-LLaMEA-gpt-4o-2",
        ]
        exp_dirs_5 = [
            "5/exp-11-01_150117-LLaMEA-gpt-4o-5",
            "5/exp-11-01_174133-LLaMEA-gpt-4o-5",
            "5/exp-11-01_192257-LLaMEA-gpt-4o-5",
        ]
        exp_dirs_10 = [
            "10/exp-11-01_150117-LLaMEA-gpt-4o-10",
            "10/exp-11-01_171429-LLaMEA-gpt-4o-10",
            "10/exp-11-01_201137-LLaMEA-gpt-4o-10",
        ]
        exp_dirs_20 = [
            "20/exp-11-01_150117-LLaMEA-gpt-4o-20",
            "20/exp-11-01_163509-LLaMEA-gpt-4o-20",
            "20/exp-11-01_200903-LLaMEA-gpt-4o-20",
        ]
        exp_dirs_40 = [
            "40/exp-11-01_150117-LLaMEA-gpt-4o-40",
            "40/exp-11-01_164611-LLaMEA-gpt-4o-40",
            "40/exp-11-01_184424-LLaMEA-gpt-4o-40",
        ]
        # TODO
        exp_dirs_var = [
            "LLaMEA/ast/dynamic_mutation_experiments/exp-gpt-4o-prompt5/exp-11-11_010046-LLaMEA-gpt-4o-beta1.5-exp1",
            "exp-11-11_025552-LLaMEA-gpt-4o-beta1.5-exp1",
            "exp-11-11_040455-LLaMEA-gpt-4o-beta1.5-exp1",
            "exp-11-11_064837-LLaMEA-gpt-4o-beta1.5-exp1",
            "exp-11-11_090052-LLaMEA-gpt-4o-beta1.5-exp1",
        ]

        colors = ["C0", "C0", "C1", "C1", "C2", "C2", "C3", "k"]
        linestyles = [
            "solid",
            "dotted",
            "solid",
            "dotted",
            "solid",
            "dotted",
            "solid",
            "solid",
        ]
        labels = ["LLaMEA-GPT4o"]
        exp_labels = ["2", "5", "10", "20", "40"]

        results = []
        for exp_i in range(
            len([exp_dirs_2, exp_dirs_5, exp_dirs_10, exp_dirs_20, exp_dirs_40])
        ):
            exp_dirs = [exp_dirs_2, exp_dirs_5, exp_dirs_10, exp_dirs_20, exp_dirs_40][
                exp_i
            ]
            exp_label = exp_labels[exp_i]

            for i in range(len(exp_dirs)):
                convergence = np.ones(budget) * -100
                # convergence_default = np.zeros(budget)
                code_diff_ratios = np.zeros(budget)
                best_so_far = -np.inf
                best_so_far_default = 0
                previous_code = ""
                previous_config = {}
                previous_name = ""
                previous_gen = 0
                parent_id = 0
                best_so_far_id = 0
                alg_id = 0
                best_ever_code = ""
                best_ever_name = ""
                best_ever_config = {}
                best_ever_fitness = -100
                log_file = base_dir + exp_dirs[i] + "/log.jsonl"
                print(log_file)
                if os.path.exists(log_file):
                    with jsonlines.open(log_file) as reader:
                        reader_i = -1
                        for obj in reader.iter(type=dict, skip_invalid=True):
                            reader_i += 1
                            stats = []
                            gen = 0
                            fitness = None
                            code_diff = 0
                            code = ""
                            if "solution" in obj.keys():
                                code = obj["solution"]
                            if "generation" in obj.keys():
                                gen = obj["generation"]
                            # if "parent_id" in obj.keys():
                            parent_id = best_so_far_id
                            if "id" in obj.keys():
                                alg_id = obj["id"]
                            if "fitness" in obj.keys():
                                fitness = obj["fitness"]
                                # print(fitness)
                            else:
                                fitness = None

                            if fitness > best_ever_fitness:
                                best_ever_fitness = fitness
                                best_ever_code = code
                                best_ever_name = obj["name"]
                                best_so_far_id = reader_i

                            if fitness <= best_so_far:
                                code_diff = code_compare(previous_code, code, False)
                            else:
                                name = obj["name"]
                                print(f"-- {gen} -- {previous_name} --> {name}")
                                code_diff = code_compare(previous_code, code, False)
                                best_so_far = fitness

                                previous_gen = gen
                                previous_code = code
                                previous_name = name

                            code_diff_ratios[gen] = code_diff
                            convergence[gen] = fitness

                            try:
                                stats = process_code(code, False)
                            except Exception as e:
                                print(e)
                                continue

                            stats["code_diff"] = code_diff
                            stats["fitness"] = 0.0
                            stats["LLM"] = labels[0] + " " + exp_label
                            stats["exp_dir"] = exp_dirs[i].replace("/", "_")
                            stats["alg_id"] = reader_i
                            # stats["gen"] = reader_i
                            stats["parent_id"] = (
                                parent_id if parent_id != None else reader_i
                            )
                            stats["fitness"] = fitness
                            results.append(stats)

        resultdf = pd.DataFrame.from_dict(results)
        print(resultdf)

        resultdf.to_csv("ast/graphstats_BBO2.csv", index=False)

    # EOH runs ------------------------------
    import pandas as pd

    base_dir = "ast/EoH_runs/"

    exp_dirs_bpo = [
        "bpo_EoHrun1",
        "bpo_EoHrun2",
        "bpo_EoHrun3",
    ]
    exp_dirs_tsp = [
        "tsp_EoHrun1",
        "tsp_EoHrun2",
        "tsp_EoHrun3",
    ]

    colors = ["C0", "C0", "C0", "C1", "C1", "C1"]
    linestyles = [
        "solid",
        "solid",
        "solid",
        "dotted",
        "dotted",
        "dotted",
    ]
    labels = ["EoH"]
    exp_labels = ["BPO", "TSP"]

    for exp_i in range(len([exp_dirs_bpo, exp_dirs_tsp])):
        results = []
        exp_dirs = [exp_dirs_bpo, exp_dirs_tsp][exp_i]
        exp_label = exp_labels[exp_i]

        for i in range(len(exp_dirs)):
            convergence = np.ones(budget) * -100
            # convergence_default = np.zeros(budget)
            code_diff_ratios = np.zeros(budget)
            best_so_far = -np.inf
            best_so_far_default = 0
            previous_code = ""
            previous_config = {}
            previous_name = ""
            previous_gen = 0
            parent_id = 0
            best_so_far_id = 0
            alg_id = 0
            best_ever_code = ""
            best_ever_name = ""
            best_ever_config = {}
            best_ever_fitness = -100
            log_file = base_dir + exp_dirs[i] + ".jsonl"
            print(log_file)
            if os.path.exists(log_file):
                with jsonlines.open(log_file) as reader:
                    reader_i = -1
                    for obj in reader.iter(type=dict, skip_invalid=True):
                        reader_i += 1
                        stats = []
                        fitness = None
                        code_diff = 0
                        code = ""
                        if "code" in obj.keys():
                            code = obj["code"]
                        if code == None:
                            continue
                        # if "parent_id" in obj.keys():
                        parents = []
                        if "id" in obj.keys():
                            alg_id = obj["id"]
                        if "parents" in obj.keys():
                            parents = obj["parents"]
                        if "objective" in obj.keys() and obj["objective"] != None:
                            fitness = obj["objective"] * -1
                            # print(fitness)
                        else:
                            fitness = -100

                        if fitness > best_ever_fitness:
                            best_ever_fitness = fitness
                            best_ever_code = code
                            best_ever_name = obj["algorithm"]
                            best_so_far_id = reader_i

                        name = obj["algorithm"]
                        # print(f"-- {gen} -- {previous_name} --> {name}")
                        best_so_far = fitness
                        previous_code = code
                        previous_name = name

                        try:
                            stats = process_code(code, False)
                        except Exception as e:
                            print(e)
                            continue

                        stats["fitness"] = 0.0
                        stats["LLM"] = labels[0] + " " + exp_label
                        stats["exp_dir"] = exp_dirs[i].replace("/", "_")
                        stats["alg_id"] = alg_id
                        stats["gen"] = reader_i
                        stats["parent_ids"] = parents
                        stats["fitness"] = fitness
                        results.append(stats)

        resultdf = pd.DataFrame.from_dict(results)
        print(resultdf)

        resultdf.to_csv(f"ast/graphstats_EOH_{exp_label}.csv", index=False)



from problems.brag_mirror import brag_mirror
from problems.ellipsometry_inverse import ellipsometry
from problems.sophisticated_antireflection_design import (
    sophisticated_antireflection_design,
)
from problems.grating2D import grating2D
from problems.plasmonic_nanostructure import plasmonic_nanostructure



# Simple helper file to plot generated auc files (used for the LLaMEA paper).
import numpy as np
import os
import seaborn as sns
import matplotlib.pyplot as plt
import copy
import tqdm
import ast
import difflib
import jellyfish


def code_compare(code1, code2):
    # Parse the Python code into ASTs
    # Use difflib to find differences
    diff = difflib.ndiff(code1.splitlines(), code2.splitlines())
    # Count the number of differing lines
    diffs = sum(1 for x in diff if x.startswith("- ") or x.startswith("+ "))
    # Calculate total lines for the ratio
    total_lines = max(len(code1.splitlines()), len(code2.splitlines()))
    similarity_ratio = (total_lines - diffs) / total_lines if total_lines else 1
    return 1 - similarity_ratio


exp_dir = "exp-05-07_14:52:37 baseline"
budget = 100

# GPT 4
exp_dirs = [
    [
        "exp-05-07_145237-gpt-4-turbo-plain",
        "exp-05-14_123400-gpt-4-turbo-plain",
        "exp-05-14_123425-gpt-4-turbo-plain",
        "exp-05-15_064710-gpt-4-turbo-plain",
        "exp-05-15_064730-gpt-4-turbo-plain",
    ],
    [
        "exp-05-13_121553-gpt-4-turbo-detail",
        "exp-05-14_123615-gpt-4-turbo-detail",
        "exp-05-14_194110-gpt-4-turbo-detail",
        "exp-05-15_064859-gpt-4-turbo-detail",
        "exp-05-17_205808-gpt-4-turbo-detail",
    ],
    [
        "exp-05-13_125616-gpt-4-turbo-elitsm",
        "exp-05-14_110804-gpt-4-turbo-elitsm",
        "exp-05-14_111732-gpt-4-turbo-elitsm",
        "exp-05-15_064825-gpt-4-turbo-elitism",
        "exp-05-15_064835-gpt-4-turbo-elitism",
    ],
    # ["exp-05-14_123738-gpt-4-turbo-detail-elitism"],
    [
        "exp-05-14_094348-gpt-3.5-turbo-plain",
        "exp-05-14_131509-gpt-3.5-turbo-plain",
        "exp-05-14_182451-gpt-3.5-turbo-plain",
        "exp-05-16_094926-gpt-3.5-turbo-plain",
        "exp-05-16_094955-gpt-3.5-turbo-plain",
    ],
    [
        "exp-05-14_094717-gpt-3.5-turbo-detail",
        "exp-05-14_192917-gpt-3.5-turbo-detail",
        "exp-05-14_182658-gpt-3.5-turbo-detail",
        "exp-05-16_095027-gpt-3.5-turbo-detail",
        "exp-05-16_095044-gpt-3.5-turbo-detail",
    ],
    [
        "exp-05-14_094824-gpt-3.5-turbo-elitism",
        "exp-05-14_182630-gpt-3.5-turbo-elitism",
        "exp-05-14_182618-gpt-3.5-turbo-elitism",
        "exp-05-16_095114-gpt-3.5-turbo-elitism",
        "exp-05-16_095135-gpt-3.5-turbo-elitism",
    ],
    [
        "exp-05-14_134318-gpt-4o-plain",
        "exp-05-16_145625-gpt-4o-plain",
        "exp-05-16_145649-gpt-4o-plain",
        "exp-05-16_145701-gpt-4o-plain",
        "exp-05-16_145709-gpt-4o-plain",
    ],
    [
        "exp-05-16_145812-gpt-4o-detail",
        "exp-05-16_145821-gpt-4o-detail",
        "exp-05-16_145835-gpt-4o-detail",
        "exp-05-16_145842-gpt-4o-detail",
        "exp-05-16_145900-gpt-4o-detail",
    ],
    [
        "exp-05-14_134351-gpt-4o-elitism",
        "exp-05-16_145906-gpt-4o-detail",
        "exp-05-16_145911-gpt-4o-detail",
        "exp-05-16_145917-gpt-4o-elitism",
        "exp-05-16_145922-gpt-4o-detail",
    ],
]

colors = [
    "b",
    "r",
    "g",
    #'c',
    "b",
    "r",
    "g",
    "b",
    "r",
    "g",
]
linestyles = [
    "solid",
    "solid",
    "solid",
    #'solid',
    "dotted",
    "dotted",
    "dotted",
    "dashed",
    "dashed",
    "dashed",
]
labels = [
    "1,1-GPT4-ES",
    "1,1 GPT4-ES w/ Details",
    "1+1 GPT4-ES",
    #'GPT-4-ES w/ Elitism+Detail',
    "1,1 GPT3.5-ES",
    "1,1 GPT3.5-ES w/ Details",
    "1+1 GPT3.5-ES",
    "1,1 GPT4o-ES",
    "1,1 GPT4o-ES w/ Details",
    "1+1 GPT4o-ES",
]
for i in range(len(exp_dirs)):
    exp_dir = exp_dirs[i]
    color = colors[i]
    ls = linestyles[i]
    label = labels[i]
    mean_aucs = []
    current_best = 0
    best_aucs = []
    std_aucs = []
    error_bars = []
    best_try = ""
    for k in range(budget):
        m_aucs = []
        for d in exp_dir:
            if os.path.isfile(f"{d}/try-{k}-aucs.txt"):
                aucs = np.loadtxt(f"{d}/try-{k}-aucs.txt")
                m_aucs.append(np.mean(aucs))
                if np.mean(aucs) > current_best:
                    current_best = np.mean(aucs)
                    best_try = f"{d}/try-{k}-aucs.txt"
        if len(m_aucs) > 0:
            mean_aucs.append(np.mean(m_aucs))
            std_aucs.append(np.std(m_aucs))
        else:
            mean_aucs.append(np.nan)
            std_aucs.append(np.nan)

        best_aucs.append(current_best)
    # print("best:",best_try,current_best)

    mean_aucs = np.array(mean_aucs)
    std_aucs = np.array(std_aucs)
    error_bars = np.array(error_bars)
    x = np.arange(budget)

    plt.plot(x, mean_aucs, color=color, linestyle=ls, label=label)
    plt.fill_between(
        x, mean_aucs - std_aucs, mean_aucs + std_aucs, color=color, alpha=0.05
    )
    # plt.fill_between(x, 0, 1, where=error_bars, color='r', alpha=0.2)
    plt.ylim(0.0, 0.7)
    plt.xlim(0, 100)
    # plt.plot(x, mean_2, 'r-', label='mean_2')
    # plt.fill_between(x, mean_2 - std_2, mean_2 + std_2, color='r', alpha=0.2)
plt.legend()
plt.tight_layout()
plt.savefig(f"comparison_plot.png")
plt.clf()


## Plot diversity for each run (difference between pop and next pop)
import re

fig = plt.figure(figsize=(8, 6))


for i in range(len(exp_dirs)):  # len(exp_dirs)
    exp_dir = exp_dirs[i]
    color = colors[i]
    label = labels[i]
    ls = linestyles[i]
    current_best = []
    elitism = False

    all_locations = []
    all_aocs = []
    if i == 2 or i == 5 or i == 8:
        elitism = True

    ratios = [[], [], [], [], []]
    previous_codes = ["", "", "", "", ""]

    if False:  # and os.path.isfile(f"ratios{i}.npy"):
        ratios = np.load(f"ratios{i}.npy")
    else:
        for k in range(budget):
            for j in range(len(exp_dir)):
                d = exp_dir[j]
                # get code file
                found = False
                for filename in os.listdir(f"{d}/code/"):
                    if re.search(f"try-{k}-\w*\.py", filename):
                        # print(filename)
                        file = open(f"{d}/code/{filename}", "r")
                        algorithm_name = re.findall(
                            "try-\d*-(\w*)\.py", filename, re.IGNORECASE
                        )[0]
                        # print(" ".join(re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', algorithm_name)), end=' ')
                        new_code = file.read()
                        file.close()
                        if k == 0:
                            ratios[j].append(0)
                        else:
                            rat = code_compare(previous_codes[j], new_code)
                            ratios[j].append(rat)
                        if os.path.isfile(f"{d}/try-{k}-aucs.txt"):
                            aucs = np.loadtxt(f"{d}/try-{k}-aucs.txt")
                            # store all code files with at least an auc

                            with open(f"algorithms/{algorithm_name}.py", "w") as file:
                                file.write(new_code)
                            while j >= len(current_best):
                                current_best.append(0)
                            if current_best[j] <= np.mean(aucs):
                                current_best[j] = np.mean(aucs)
                                if elitism:
                                    previous_codes[j] = new_code
                            all_aocs.append(np.mean(aucs))
                            all_locations.append(f"{d}/code/{filename}")
                        if not elitism:
                            previous_codes[j] = new_code
                        found = True
                        break
                if not found:
                    ratios[j].append(0)
        ratios = np.array(ratios)
        mean_ratios = np.mean(ratios, axis=0)
        # std_aucs = np.array(std_aucs)
        # np.save(f"ratios{i}.npy", ratios)

    mean_ratios = np.mean(ratios, axis=0)
    x = np.arange(budget)
    for j in range(ratios.shape[0]):
        plt.plot(x, ratios[j], color=color, linestyle=ls, alpha=0.02)
    plt.plot(x, mean_ratios, color=color, linestyle=ls, label=label)
    # plt.fill_between(x, mean_aucs - std_aucs, mean_aucs + std_aucs, color=color, alpha=0.05)
    # plt.fill_between(x, 0, 1, where=error_bars, color='r', alpha=0.2)

    # best 3 algorithms per run
    # all_aocs = np.array(all_aocs)
    # all_locations = np.array(all_locations)
    # indices_best = np.argsort(all_aocs)[-3:]
    # print(all_locations[indices_best], all_aocs[indices_best])

    # plt.plot(x, mean_2, 'r-', label='mean_2')
    # plt.fill_between(x, mean_2 - std_2, mean_2 + std_2, color='r', alpha=0.2)
# plt.yscale("log")
# plt.xscale("log")
plt.ylim(0.0, 1.0)
plt.xlim(0.0, 100)
plt.ylabel(
    "pairwise difference ratio"
)  # Bold lines depict averages over all runs, more transparent lines depict independent runs.
plt.xlabel("Iterations")
plt.legend()
plt.tight_layout()

plt.savefig(f"diff_plot0.pdf")
plt.clf()


# diff names
fig = plt.figure(figsize=(8, 4))


for i in range(len(exp_dirs)):  # len(exp_dirs)
    exp_dir = exp_dirs[i]
    color = colors[i]
    label = labels[i]
    ls = linestyles[i]
    current_best = []
    current_best_name = []
    elitism = False

    all_locations = []
    all_aocs = []
    if i == 2 or i == 5 or i == 8:
        elitism = True

    ratios = [[], [], [], [], []]
    ratios_name = [[], [], [], [], []]
    previous_codes = ["", "", "", "", ""]
    previous_names = ["", "", "", "", ""]

    if False:  # and os.path.isfile(f"ratios{i}.npy"):
        ratios = np.load(f"ratios{i}.npy")
    else:
        for k in range(budget):
            for j in range(len(exp_dir)):
                d = exp_dir[j]
                # get code file
                found = False
                for filename in os.listdir(f"{d}/code/"):
                    if re.search(f"try-{k}-\w*\.py", filename):
                        # print(filename)
                        algorithm_name = re.findall(
                            "try-\d*-(\w*)\.py", filename, re.IGNORECASE
                        )[0]
                        if k == 0:
                            ratios_name[j].append(1)
                        else:
                            rat2 = jellyfish.jaro_similarity(
                                previous_names[j], algorithm_name
                            )
                            ratios[j].append(rat)
                            ratios_name[j].append(rat2)
                        if os.path.isfile(f"{d}/try-{k}-aucs.txt"):
                            aucs = np.loadtxt(f"{d}/try-{k}-aucs.txt")
                            # store all code files with at least an auc

                            # with open(f"algorithms/{algorithm_name}.py", "w") as file:
                            #     file.write(new_code)
                            while j >= len(current_best):
                                current_best.append(0)
                            if current_best[j] <= np.mean(aucs):
                                current_best[j] = np.mean(aucs)
                                if elitism:
                                    previous_codes[j] = new_code
                                    previous_names[j] = algorithm_name
                            all_aocs.append(np.mean(aucs))
                        if not elitism:
                            previous_codes[j] = new_code
                            previous_names[j] = algorithm_name
                        found = True
                        break
                if not found:
                    ratios_name[j].append(0)
        print(ratios_name)
        ratios_name = np.array(ratios_name)
        mean_ratios_name = np.mean(ratios_name, axis=0)

    mean_ratios_name = np.mean(ratios_name, axis=0)
    x = np.arange(budget)
    # for j in range(ratios_name.shape[0]):
    #    plt.plot(x, ratios_name[j], color=color, linestyle=ls, alpha=0.02)
    plt.plot(x, mean_ratios_name, color=color, linestyle=ls, label=label)
    # plt.fill_between(x, mean_aucs - std_aucs, mean_aucs + std_aucs, color=color, alpha=0.05)
    # plt.fill_between(x, 0, 1, where=error_bars, color='r', alpha=0.2)

    # best 3 algorithms per run
    # all_aocs = np.array(all_aocs)
    # all_locations = np.array(all_locations)
    # indices_best = np.argsort(all_aocs)[-3:]
    # print(all_locations[indices_best], all_aocs[indices_best])

    # plt.plot(x, mean_2, 'r-', label='mean_2')
    # plt.fill_between(x, mean_2 - std_2, mean_2 + std_2, color='r', alpha=0.2)
# plt.yscale("log")
# plt.xscale("log")
plt.ylim(0.0, 1.0)
plt.xlim(0.0, 100)
plt.ylabel(
    "Pairwise Jaro similarity"
)  # Bold lines depict averages over all runs, more transparent lines depict independent runs.
plt.xlabel("Iterations")
plt.legend()
plt.tight_layout()

plt.savefig(f"diff_plot_names.pdf")
plt.clf()


# diff names independent runs
# diff names
fig = plt.figure(figsize=(6, 4))


for i in range(len(exp_dirs)):  # len(exp_dirs)
    exp_dir = exp_dirs[i]
    color = colors[i]
    label = labels[i]
    ls = linestyles[i]
    current_best = []
    current_best_name = []
    elitism = False

    all_locations = []
    all_aocs = []
    if i == 2 or i == 5 or i == 8:
        elitism = True

    ratios = [[], [], [], [], []]
    ratios_name = [[], [], [], [], []]
    previous_codes = ["", "", "", "", ""]
    previous_names = ["", "", "", "", ""]

    if False:  # and os.path.isfile(f"ratios{i}.npy"):
        ratios = np.load(f"ratios{i}.npy")
    else:
        for k in range(budget):
            for j in range(len(exp_dir)):
                d = exp_dir[j]
                # get code file
                found = False
                for filename in os.listdir(f"{d}/code/"):
                    if re.search(f"try-{k}-\w*\.py", filename):
                        # print(filename)
                        algorithm_name = re.findall(
                            "try-\d*-(\w*)\.py", filename, re.IGNORECASE
                        )[0]
                        if k == 0:
                            ratios_name[j].append(1)
                        else:
                            rat2 = jellyfish.jaro_similarity(
                                previous_names[j], algorithm_name
                            )
                            ratios[j].append(rat)
                            ratios_name[j].append(rat2)
                        if os.path.isfile(f"{d}/try-{k}-aucs.txt"):
                            aucs = np.loadtxt(f"{d}/try-{k}-aucs.txt")
                            # store all code files with at least an auc

                            # with open(f"algorithms/{algorithm_name}.py", "w") as file:
                            #     file.write(new_code)
                            while j >= len(current_best):
                                current_best.append(0)
                            if current_best[j] <= np.mean(aucs):
                                current_best[j] = np.mean(aucs)
                                if elitism:
                                    previous_codes[j] = new_code
                                    previous_names[j] = algorithm_name
                            all_aocs.append(np.mean(aucs))
                        if not elitism:
                            previous_codes[j] = new_code
                            previous_names[j] = algorithm_name
                        found = True
                        break
                if not found:
                    ratios_name[j].append(0)
        print(ratios_name)
        ratios_name = np.array(ratios_name)
        mean_ratios_name = np.mean(ratios_name, axis=0)

    mean_ratios_name = np.mean(ratios_name, axis=0)
    x = np.arange(budget)
    for j in range(ratios_name.shape[0]):
        if j == 0:
            plt.plot(x, ratios_name[j], color=color, linestyle=ls, label=label)
        else:
            plt.plot(x, ratios_name[j], color=color, linestyle=ls)
    # plt.plot(x, mean_ratios_name, color=color, linestyle=ls, label=label)
    # plt.fill_between(x, mean_aucs - std_aucs, mean_aucs + std_aucs, color=color, alpha=0.05)
    # plt.fill_between(x, 0, 1, where=error_bars, color='r', alpha=0.2)

    # best 3 algorithms per run
    # all_aocs = np.array(all_aocs)
    # all_locations = np.array(all_locations)
    # indices_best = np.argsort(all_aocs)[-3:]
    # print(all_locations[indices_best], all_aocs[indices_best])

    # plt.plot(x, mean_2, 'r-', label='mean_2')
    # plt.fill_between(x, mean_2 - std_2, mean_2 + std_2, color='r', alpha=0.2)
    # plt.yscale("log")
    # plt.xscale("log")
    plt.ylim(0.0, 1.0)
    plt.xlim(0.0, 100)
    plt.ylabel(
        "Pairwise Jaro similarity"
    )  # Bold lines depict averages over all runs, more transparent lines depict independent runs.
    plt.xlabel("Iterations")
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"diff_plot_names_{i}.pdf")
    plt.clf()

# GPT 4
fig = plt.figure(figsize=(8, 6))
# exp_dirs = [["exp-05-07_14:52:37 baseline","exp-05-14_12:34:00-gpt-4-turbo-plain", "exp-05-14_12:34:25-gpt-4-turbo-plain"], ["exp-05-13_12:15:53 detailed", "exp-05-14_12:36:03-gpt-4-turbo-detail", "exp-05-14_12:36:15-gpt-4-turbo-detail"], ["exp-05-13_12:56:16 elitism", "exp-05-14_11:08:04-gpt-4-turbo-elitsm"], ["exp-05-14_12:37:38-gpt-4-turbo-detail-elitism"]
#     , ["exp-05-14_09:43:48"], ["exp-05-14_09:47:17"], ["exp-05-14_09:48:24"]]
# colors = ['b','r','g','k', 'c', 'm', 'y']
# labels = ['GPT-4-ES', 'GPT-4-ES w/ Detail', 'GPT-4-ES w/ Elitism', 'GPT-4-ES w/ Elitism+Detail', 'GPT-3.5-ES', 'GPT-3.5-ES w/ Details', 'GPT-3.5-ES w/ Elitism']
for i in range(len(exp_dirs)):
    exp_dir = exp_dirs[i]
    color = colors[i]
    label = labels[i]
    ls = linestyles[i]
    mean_aucs = []
    current_best = []
    best_aucs = []
    std_aucs = []
    error_bars = []
    best_try = ""
    for k in range(budget):
        m_aucs = []
        for j in range(len(exp_dir)):
            d = exp_dir[j]
            if os.path.isfile(f"{d}/try-{k}-aucs.txt"):
                aucs = np.loadtxt(f"{d}/try-{k}-aucs.txt")
                m_aucs.append(np.mean(aucs))
                while j >= len(current_best):
                    current_best.append(0)
                if current_best[j] < np.mean(aucs):
                    current_best[j] = np.mean(aucs)
        if len(current_best) > 0:
            mean_aucs.append(np.mean(current_best))
            std_aucs.append(np.std(current_best))
        else:
            mean_aucs.append(np.nan)
            std_aucs.append(np.nan)

    mean_aucs = np.array(mean_aucs)
    std_aucs = np.array(std_aucs)
    x = np.arange(budget)

    plt.plot(x, mean_aucs, color=color, linestyle=ls, label=label)
    plt.fill_between(
        x, mean_aucs - std_aucs, mean_aucs + std_aucs, color=color, alpha=0.05
    )
    # plt.fill_between(x, 0, 1, where=error_bars, color='r', alpha=0.2)

    # plt.plot(x, mean_2, 'r-', label='mean_2')
    # plt.fill_between(x, mean_2 - std_2, mean_2 + std_2, color='r', alpha=0.2)
# plt.yscale("symlog")
# plt.xscale("symlog")
plt.ylim(0.0, 0.6)
plt.xlim(0.0, 100)
plt.ylabel("mean AOCC")
plt.xlabel("Iterations")
plt.legend()
plt.tight_layout()

plt.savefig(f"convergence_plot.pdf")
plt.clf()
# plt.plot(x, best_aucs, 'b-', label='baseline')
# plt.ylim(0.0,1.0)
# #plt.plot(x, mean_2, 'r-', label='mean_2')
# #plt.fill_between(x, mean_2 - std_2, mean_2 + std_2, color='r', alpha=0.2)
# plt.legend()
# plt.tight_layout()
# plt.savefig(f"{exp_dir}/convergence_plot.png")



# This is a more advanced example of how to use the LLaMEA algorithm with the Gemini LLM to generate optimization algorithms for the BBOB test suite.
# It includes the in-the-loop hyper-parameter optimization (HPO) using SMAC extension of LLaMEA and a more complex evaluation function.
# We have to define the following components for LLaMEA to work:
# - An evaluation function that executes the generated code and evaluates its performance.
# - A task prompt that describes the problem to be solved.
# - An LLM instance that will generate the code based on the task prompt.


import os
import re
import time
from itertools import product

import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace
from ioh import get_problem, logger
from smac import AlgorithmConfigurationFacade, Scenario

from llamea import Gemini_LLM, LLaMEA
from misc import OverBudgetException, aoc_logger, correct_aoc

if __name__ == "__main__":
    # Execution code starts here
    api_key = os.getenv("GEMINI_API_KEY")
    ai_model = "gemini-1.5-flash"
    experiment_name = "pop1-5-HPO"
    llm = Gemini_LLM(api_key, ai_model)

    def evaluateBBOBWithHPO(solution, explogger=None):
        """
        Evaluates an optimization algorithm on the BBOB (Black-Box Optimization Benchmarking) suite and computes
        the Area Over the Convergence Curve (AOCC) to measure performance. In addddition, if a configuration space is provided, it
        applies Hyper-parameter optimization with SMAC first.

        Parameters:
        -----------
        solution : dict
            A dictionary containing "_solution" (the code to evaluate), "_name", "_description" and "_configspace"

        explogger : logger
            A class to log additional stuff for the experiment.

        Returns:
        --------
        solution : dict
            Updated solution with "_fitness", "_feedback", "incumbent" and optional "_error"

        Functionality:
        --------------
        - Executes the provided `code` string in the global context, allowing for dynamic inclusion of necessary components.
        - Iterates over a predefined set of dimensions (currently only 5), function IDs (1 to 24), and instance IDs (1 to 3).
        - For each problem, the specified algorithm is instantiated and executed with a defined budget.
        - AOCC is computed for each run, and the results are aggregated across all runs, problems, and repetitions.
        - The function handles cases where the algorithm exceeds its budget using an `OverBudgetException`.
        - Logs the results if an `explogger` is provided.
        - The function returns a feedback string, the mean AOCC score, and an error placeholder.

        Notes:
        ------
        - The budget for each algorithm run is set to 10,000.
        - The function currently only evaluates a single dimension (5), but this can be extended.
        - Hyperparameter Optimization (HPO) with SMAC is mentioned but not implemented.
        - The AOCC score is a metric where 1.0 is the best possible outcome, indicating optimal convergence.

        """
        auc_mean = 0
        auc_std = 0
        code = solution.code
        algorithm_name = solution.name
        exec(code, globals())
        dim = 5
        budget = 2000 * dim
        error = ""
        algorithm = None

        # perform a small run to check for any code errors
        l2_temp = aoc_logger(100, upper=1e2, triggers=[logger.trigger.ALWAYS])
        problem = get_problem(11, 1, dim)
        problem.attach_logger(l2_temp)
        try:
            algorithm = globals()[algorithm_name](budget=100, dim=dim)
            algorithm(problem)
        except OverBudgetException:
            pass

        # now optimize the hyper-parameters
        def get_bbob_performance(config: Configuration, instance: str, seed: int = 0):
            np.random.seed(seed)
            fid, iid = instance.split(",")
            fid = int(fid[1:])
            iid = int(iid[:-1])
            problem = get_problem(fid, iid, dim)
            l2 = aoc_logger(budget, upper=1e2, triggers=[logger.trigger.ALWAYS])
            problem.attach_logger(l2)
            try:
                algorithm = globals()[algorithm_name](
                    budget=budget, dim=dim, **dict(config)
                )
                algorithm(problem)
            except OverBudgetException:
                pass
            except Exception as e:
                print(problem.state, budget, e)
            auc = correct_aoc(problem, l2, budget)
            return 1 - auc

        args = list(product(range(1, 25), range(1, 4)))
        np.random.shuffle(args)
        inst_feats = {str(arg): [arg[0]] for idx, arg in enumerate(args)}
        # inst_feats = {str(arg): [idx] for idx, arg in enumerate(args)}
        error = ""

        if "_configspace" not in solution.keys():
            # No HPO possible, evaluate only the default
            incumbent = {}
            error = "The configuration space was not properly formatted or not present in your answer. The evaluation was done on the default configuration."
        else:
            configuration_space = solution.configspace
            scenario = Scenario(
                configuration_space,
                name=str(int(time.time())) + "-" + algorithm_name,
                deterministic=False,
                min_budget=12,
                max_budget=200,
                n_trials=2000,
                instances=args,
                instance_features=inst_feats,
                output_directory="smac3_output"
                if explogger is None
                else explogger.dirname + "/smac"
                # n_workers=10
            )
            smac = AlgorithmConfigurationFacade(
                scenario, get_bbob_performance, logging_level=30
            )
            incumbent = smac.optimize()

        # last but not least, perform the final validation

        l2 = aoc_logger(budget, upper=1e2, triggers=[logger.trigger.ALWAYS])
        aucs = []
        for fid in np.arange(1, 25):
            for iid in [1, 2, 3]:  # , 4, 5]
                problem = get_problem(fid, iid, dim)
                problem.attach_logger(l2)
                for rep in range(3):
                    np.random.seed(rep)
                    try:
                        algorithm = globals()[algorithm_name](
                            budget=budget, dim=dim, **dict(incumbent)
                        )
                        algorithm(problem)
                    except OverBudgetException:
                        pass
                    auc = correct_aoc(problem, l2, budget)
                    aucs.append(auc)
                    l2.reset(problem)
                    problem.reset()

        auc_mean = np.mean(aucs)
        auc_std = np.std(aucs)
        dict_hyperparams = dict(incumbent)
        feedback = f"The algorithm {algorithm_name} got an average Area over the convergence curve (AOCC, 1.0 is the best) score of {auc_mean:0.2f} with optimal hyperparameters {dict_hyperparams}."
        print(algorithm_name, algorithm, auc_mean, auc_std)

        solution.add_metadata("aucs", aucs)
        solution.add_metadata("incumbent", dict_hyperparams)
        solution.set_scores(auc_mean, feedback)

        return solution

    role_prompt = "You are a highly skilled computer scientist in the field of natural computing. Your task is to design novel metaheuristic algorithms to solve black box optimization problems."
    task_prompt = """
    The optimization algorithm should handle a wide range of tasks, which is evaluated on the BBOB test suite of 24 noiseless functions. Your task is to write the optimization algorithm in Python code. The code should contain an `__init__(self, budget, dim)` function with optional additional arguments and the function `def __call__(self, func)`, which should optimize the black box function `func` using `self.budget` function evaluations.
    The func() can only be called as many times as the budget allows, not more. Each of the optimization functions has a search space between -5.0 (lower bound) and 5.0 (upper bound). The dimensionality can be varied.

    In addition, any hyper-parameters the algorithm uses will be optimized by SMAC, for this, provide a Configuration space as Python dictionary (without the dim and budget parameters) and include all hyper-parameters in the __init__ function header.
    An example configuration space is as follows:

    ```python
    {
        "float_parameter": (0.1, 1.5),
        "int_parameter": (2, 10), 
        "categoral_parameter": ["mouse", "cat", "dog"]
    }
    ```

    Give an excellent and novel heuristic algorithm including its configuration space to solve this task and also give it a one-line description, describing the main idea. 
    """

    format_prompt = """
    Give the response in the format:
    # Description: <short-description>
    # Code: <code>
    # Space: <configuration_space>
    """

    example_prompt = """
    An example of such code (a simple random search), is as follows:
    ```python
    import numpy as np

    class RandomSearch:
        def __init__(self, budget=10000, dim=10):
            self.budget = budget
            self.dim = dim

        def __call__(self, func):
            self.f_opt = np.inf
            self.x_opt = None
            for i in range(self.budget):
                x = np.random.uniform(func.bounds.lb, func.bounds.ub)
                
                f = func(x)
                if f < self.f_opt:
                    self.f_opt = f
                    self.x_opt = x
                
            return self.f_opt, self.x_opt
    ```
    """

    feedback_prompts = [
        f"Either refine or redesign to improve the solution (and give it a distinct one-line description)."
    ]

    for experiment_i in [1]:
        es = LLaMEA(
            evaluateBBOBWithHPO,
            llm=llm,
            role_prompt=role_prompt,
            task_prompt=task_prompt,
            example_prompt=example_prompt,
            output_format_prompt=format_prompt,
            mutation_prompts=feedback_prompts,
            experiment_name=experiment_name,
            elitism=True,
            HPO=True,
        )
        print(es.run())



import numpy as np
import re


try:
    from ioh import LogInfo
    # try:
    #     from ioh.logger import AbstractLogger
    # except ImportError:
    from ioh import logger as iohLogger
    AbstractLogger = iohLogger.AbstractLogger
except ImportError:
    LogInfo = None
    AbstractLogger = object




class ThresholdReachedException(Exception):
    """The algorithm reached the lower threshold."""

    pass


class OverBudgetException(Exception):
    """The algorithm tried to do more evaluations than allowed."""

    pass


def correct_aoc(ioh_function, logger, budget):
    """Correct aoc values in case a run stopped before the budget was exhausted

    Args:
        ioh_function: The function in its final state (before resetting!)
        logger: The logger in its final state, so we can ensure the settings for aoc calculation match
        budget: The intended maximum budget

    Returns:
        float: The normalized aoc of the run, corrected for stopped runs
    """
    fraction = (
        logger.transform(
            np.clip(
                ioh_function.state.current_best_internal.y, logger.lower, logger.upper
            )
        )
        - logger.transform(logger.lower)
    ) / (logger.transform(logger.upper) - logger.transform(logger.lower))
    aoc = (
        logger.aoc
        + np.clip(budget - ioh_function.state.evaluations, 0, budget) * fraction
    ) / budget

    return 1 - aoc


class aoc_logger(AbstractLogger):
    """aoc_logger class implementing the logging module for ioh."""

    def __init__(
        self,
        budget,
        lower=1e-8,
        upper=1e8,
        scale_log=True,
        stop_on_threshold=False,
        *args,
        **kwargs,
    ):
        """Initialize the logger.

        Args:
            budget (int): Evaluation budget for calculating aoc.
        """
        super().__init__(*args, **kwargs)
        self.aoc = 0
        self.lower = lower
        self.upper = upper
        self.budget = budget
        self.stop_on_threshold = stop_on_threshold
        self.transform = lambda x: np.log10(x) if scale_log else (lambda x: x)

    def __call__(self, log_info: LogInfo):
        """Subscalculate the aoc.

        Args:
            log_info (ioh.LogInfo): info about current values.
        """
        if log_info.evaluations > self.budget:
            raise OverBudgetException
        if log_info.evaluations == self.budget:
            return
        if self.stop_on_threshold and abs(log_info.raw_y_best) < self.lower:
            raise ThresholdReachedException
        y_value = np.clip(log_info.raw_y_best, self.lower, self.upper)
        self.aoc += (self.transform(y_value) - self.transform(self.lower)) / (
            self.transform(self.upper) - self.transform(self.lower)
        )

    def reset(self, func):
        super().reset()
        self.aoc = 0


class budget_logger(AbstractLogger):
    """budget_logger class implementing the logging module for ioh."""

    def __init__(
        self,
        budget,
        *args,
        **kwargs,
    ):
        """Initialize the logger.

        Args:
            budget (int): Evaluation budget for calculating aoc.
        """
        super().__init__(*args, **kwargs)
        self.budget = budget

    def __call__(self, log_info: LogInfo):
        """Subscalculate the aoc.

        Args:
            log_info (ioh.LogInfo): info about current values.
        """
        if log_info.evaluations > self.budget:
            raise OverBudgetException

    def reset(self):
        super().reset()





import re
import os
import ast
import importlib
import jsonlines
import numpy as np
from typing import Any, Optional
from difflib import SequenceMatcher


class NoCodeException(Exception):
    """Could not extract generated code."""

    pass


def handle_timeout(signum, frame):
    """Raise a timeout exception"""
    raise TimeoutError


def _code_updater(code: str, lines_to_change: list[str], updated_lines: list[str]):
    """Line by line update code, and return the update.
    Args:
        code: Current code in the individual.
        lines_to_change: A list of lines to be changed by the LLM.
        updated_lines: Lines to replace the `lines_to_update`.

    """
    if len(lines_to_change) != len(lines_to_change):
        raise ValueError
    for i in range(len(lines_to_change)):
        code = code.replace(
            lines_to_change[i], updated_lines[i], 1
        )  # Update one occurance of lines_to_change, to corresponding change.
    return code


def apply_code_delta(text: str, base_code: str) -> tuple[str, bool, float]:
    """
    Assuming the LLM follows the intructions properly, following format of response is expected.
    ```diff <- (diff may appear sometimes.)
    # A series of following search replace pattern will appear.
    <<<<<<< SEARCH
    for i in range(m):
        for j in range(p):
            for k in range(n):
                C[i, j] += A[i, k] * B[k, j]
    =======
    # Reorder loops for better memory access pattern
    for i in range(m):
        for k in range(n):
            for j in range(p):
                C[i, j] += A[i, k] * B[k, j]
    >>>>>>> REPLACE
    ```

    Args:
        text: LLM response.text.
        base_code: Base code to be mutated.
    Returns:
        Code: updated code, after applying diff.
        bool: Success of diff mode implementation.
        float: Ratio of new code similar to the original `base_code`.
    """
    outLines = []
    inLines = []
    try:
        pattern = re.compile(
            r"(?s)<{3,}\s*SEARCH\s*\n(.*?)\n={3,}\s*\n(.*?)(?=\n>{3,}\s*REPLACE)"
        )
        matches = pattern.findall(text)
        if len(matches) == 0:
            print(
                "WARNING: LLM didn't adhere to search replace pattern. Try bigger model."
            )
            raise ValueError

        for search, replace in matches:
            outLines.append(search)
            inLines.append(replace)

        code = _code_updater(base_code, outLines, inLines)

        seq_match = SequenceMatcher(None, code, base_code)
        ratio = seq_match.ratio()

        return code, True, ratio

    except Exception:
        return base_code, False, 1.0


def discrete_power_law_distribution(n, beta):
    """
    Power law distribution function from:
    # Benjamin Doerr, Huu Phuoc Le, Régis Makhmara, and Ta Duy Nguyen. 2017.
    # Fast genetic algorithms.
    # In Proceedings of the Genetic and Evolutionary Computation Conference (GECCO '17).
    # Association for Computing Machinery, New York, NY, USA, 777–784.
    # https://doi.org/10.1145/3071178.3071301
    """

    def discrete_power_law(n, alpha, beta):
        half_n = int(n / 2)
        C_beta_half_n = 0
        for i in range(1, half_n + 1):
            C_beta_half_n += i ** (-beta)
        probability_alpha = C_beta_half_n ** (-1) * alpha ** (-beta)
        return probability_alpha

    half_n = int(n / 2)
    elements = [alpha for alpha in range(1, half_n + 1)]
    probabilities = [discrete_power_law(n, alpha, beta) for alpha in elements]
    if elements == []:
        return 0.05
    else:
        sample = np.random.choice(elements, p=probabilities)
        return sample / n


def code_distance(a, b):
    """Return a rough distance between two solutions based on their ASTs.

    The function accepts either :class:`Solution` objects or raw code strings
    and computes ``1 - similarity`` of their abstract syntax trees using
    :class:`difflib.SequenceMatcher` on the dumped AST representations.
    ``1.0`` is returned on parsing errors or when the inputs cannot be
    processed.

    Args:
        a: The first solution or Python source code.
        b: The second solution or Python source code.

    Returns:
        float: A value in ``[0, 1]`` indicating dissimilarity of the code.
    """

    code_a = getattr(a, "code", a)
    code_b = getattr(b, "code", b)
    try:
        tree_a = ast.parse(code_a)
        tree_b = ast.parse(code_b)
        return 1 - SequenceMatcher(None, ast.dump(tree_a), ast.dump(tree_b)).ratio()
    except Exception:
        return 1.0


def _collect_imports(code: str):
    """Collect import info from code using AST.

    Args:
        `code: str` The source code as a string.
    Returns:
        `imports: [{str: str | None}]`: A list of import symbols, containing import type "from" | "import",
            module name in "module", sub module name in "name", and alias name-followed by `as` keyword--in "alias".
    """
    tree = ast.parse(code)
    imports = []

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imports.append(
                    {"type": "import", "module": alias.name, "alias": alias.asname}
                )
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                imports.append(
                    {
                        "type": "from",
                        "module": node.module,
                        "name": alias.name,
                        "alias": alias.asname,
                    }
                )
    return imports


def _add_builtins_into(allowed_list: list[str]):
    """
    Adds safe `__builtins__` library to allowed_list.

    Args:
        `allowed_list: list[str]: ` A list of allowed libraries, that are pip installable.

    Returns:
        `None` (Uses reference semantics to add `__builtins__` to `allowed_list`).
    """
    allowed_list += ["math", "random", "statistics", "itertools", "operator", "heapq"]


def prepare_namespace(
    code: str, allowed: list[str], logger: Any = None
) -> tuple[dict[str, Any], Optional[str]]:
    """Prepare exec global_namespace, with the libraries imported in the text, `code` parameter accepts.
        If the imports are not allowed in the environment, a generic object is provided.

    ### Args:
        `code: str`: Code parameter that is to be passed to `exec` function.

        `allowed: list[str]`: A list of allowed pip installable libraries, that are acceptable to be imported.

        `logger: Any`: Logger with `log_import_fail(list[str])` method in it, LLaMEA has this feature in llamea.loggers.ExperimentLogger.

    ### Returns:
        Returns a prepared global_namespace dictionary for exec, of type `dict[str, Any]`, along with an str,
        `potential_issue`, which can be passed out to feedback to LLM when `exec` throws.

    """
    ns = {}
    imports = _collect_imports(code)

    allowed = allowed.copy()
    allowed = list(map(lambda x: x.split(">")[0], allowed))
    _add_builtins_into(allowed)
    not_allowed: list[str] = []

    for imp in imports:
        if imp["type"] == "import":
            module = imp["module"]

            if allowed and not any(
                module == a or module.startswith(a + ".") for a in allowed
            ):
                ns[imp["alias"] or module.split(".")[0]] = object
                not_allowed.append(imp["module"])
            else:
                mod = importlib.import_module(module)
                ns[imp["alias"] or module.split(".")[0]] = mod

        elif imp["type"] == "from":
            module = imp["module"]

            if allowed and not any(
                module == a or module.startswith(a + ".") for a in allowed
            ):
                ns[imp["alias"] or imp["name"]] = object
                not_allowed.append(imp["module"])
            else:
                mod = importlib.import_module(module)
                obj = getattr(mod, imp["name"])
                ns[imp["alias"] or imp["name"]] = obj

    potential_issue = None

    if logger:
        try:
            logger.log_import_fails(not_allowed)
        except Exception as e:
            print("Provided logger doesn't have log_import_fail", e.__repr__())

    if len(not_allowed) > 0:
        potential_issue = (
            ", ".join(not_allowed)
            + f" {'are' if len(not_allowed) > 1 else 'is'} currently not allowed to be imported in this framework."
        )
    return (ns, potential_issue)


def clean_local_namespace(
    local_namespace: dict[str, Any], global_namespace: dict[str, Any]
):
    """The exec command upon execution, adds global_namespace parameters to local_namespace parameters.
    This function returns local_ns - gobal_ns, so that sweeping for object type never returns a library imported objects.

    Args:
        `local_namespace : dict[str, Any]`: Dictionary that was passed as local_namespace to `exec` block.
        `global_namespace : dict[str, Any]`: Dictionary/Mapping passed as global_namespace to `exec` block.

    Returns:
        Original `local_namespace`, that is `local_namespace` - `global_namespace`.
    """
    for key in global_namespace:
        if key in local_namespace:
            local_namespace.pop(key)
    return local_namespace



import os
import time
import jsonlines

from llamea.loggers import ExperimentLogger
from llamea.solution import Solution

soln = Solution(
            code="""
import numpy as np
from scipy.spatial import ConvexHull
import random
import math
from statistics import mean

class HeilbronnConvexRegion_n14:

    def __init__(self, n_points: int, best_known_configuration: list[float] | None = None):
        self.n_points = n_points
        self.best_known_configuration = best_known_configuration
        self.points = None
        if best_known_configuration is not None:
            self.points = np.array(best_known_configuration).reshape((n_points, 2))

    def _area(self, points):
        hull = ConvexHull(points)
        return hull.area

    def _rescale(self, points):
        area = self._area(points)
        scale = math.sqrt(1.0 / area)
        centroid = np.mean(points, axis=0)
        points = (points - centroid) * scale + centroid
        return points

    def _smallest_triangle_area(self, points):
        min_area = float('inf')
        for i in range(self.n_points):
            for j in range(i + 1, self.n_points):
                for k in range(j + 1, self.n_points):
                    area = 0.5 * abs(
                        points[i, 0] * (points[j, 1] - points[k, 1]) +
                        points[j, 0] * (points[k, 1] - points[i, 1]) +
                        points[k, 0] * (points[i, 1] - points[j, 1])
                    )
                    min_area = min(min_area, area)
        return min_area

    def _generate_initial_points(self):
        return np.random.rand(self.n_points, 2)

    def __call__(self):
        if self.points is None:
            self.points = self._generate_initial_points()

        self.points = self._rescale(self.points)

        best_points = self.points.copy()
        best_score = self._smallest_triangle_area(self.points)

        temperature = 1.0
        cooling_rate = 0.995
        tolerance = 1e-12

        for _ in range(100000):
            i = random.randint(0, self.n_points - 1)
            original_point = self.points[i].copy()

            # Perturb the point
            self.points[i] += np.random.normal(0, 0.01, 2)

            # Rescale to maintain area of 1
            self.points = self._rescale(self.points)

            new_score = self._smallest_triangle_area(self.points)

            if new_score > best_score:
                best_score = new_score
                best_points = self.points.copy()
            else:
                # Simulated annealing acceptance probability
                acceptance_probability = math.exp((new_score - best_score) / temperature)
                if random.random() < acceptance_probability:
                    pass  # Accept the worse solution
                else:
                    self.points[i] = original_point  # Revert the change

            temperature *= cooling_rate

            if temperature < tolerance:
                break

        return best_points
""",
        name="HeilbronnConvexRegion_n14",
        generation=4
        )

def test_logger_logs_solution():
    solution = soln.copy()
    solution.set_scores(
        0.02783557145848216,
        "min_triangle_area=0.0278356, best known = 0.0278."
    )
    test = {
        "solution" : solution,
        "output" : [[0.63557142, 0.2780757 ],
                    [0.41179303, 0.6500378 ],
                    [0.71410244, 0.29145972],
                    [0.35526134, 0.29541504],
                    [0.38375729, 0.3046403 ],
                    [0.55842241, 0.7181737 ],
                    [0.53782542, 0.32973947],
                    [0.67912473, 0.6429485 ],
                    [0.59160907, 0.55614398],
                    [0.27554739, 0.71103061],
                    [0.48850259, 0.26512655],
                    [0.79653191, 0.71262127],
                    [0.51986098, 0.77731033],
                    [0.75680552, 0.64723116]
                    ]
    }
    logger = ExperimentLogger("test_soln_logging")
    logger.log_solution(test["solution"], test["output"])
    
    last_element = {}
    time.sleep(1)

    with jsonlines.open(os.path.join(logger.dirname, "solutions.jsonl"), "r") as reader:
        for line in reader:
            last_element = line
    
    assert last_element["id"] == test["solution"].id
    assert last_element["fitness"] == test["solution"].fitness
    assert last_element["output"] == test["output"]

def test_logger_ignores_bad_fitness_solutions():
    # Test solution with missing fitness:
    # soln.set_scores(                                                      #Forgot using setscore before logging.
    #     0.02783557145848216,
    #     "min_triangle_area=0.0278356, best known = 0.0278."
    # )


    test = {
        "solution" : soln.copy(),
        "output" : [[0.63557142, 0.2780757 ],
                    [0.41179303, 0.6500378 ],
                    [0.71410244, 0.29145972],
                    [0.35526134, 0.29541504],
                    [0.38375729, 0.3046403 ],
                    [0.55842241, 0.7181737 ],
                    [0.53782542, 0.32973947],
                    [0.67912473, 0.6429485 ],
                    [0.59160907, 0.55614398],
                    [0.27554739, 0.71103061],
                    [0.48850259, 0.26512655],
                    [0.79653191, 0.71262127],
                    [0.51986098, 0.77731033],
                    [0.75680552, 0.64723116]
                    ]
    }
    logger = ExperimentLogger("test_logger_ignores_bad_soln")
    logger.log_solution(test["solution"], test["output"])
    
    last_element = {}
    time.sleep(1)

    with jsonlines.open(os.path.join(logger.dirname, "solutions.jsonl"), "r") as reader:
        for line in reader:
            last_element = line
    
    assert last_element == {}

def test_logger_ignore_solution_without_code():
    solution = soln.copy()
    solution.code = ""

    solution.set_scores(
        0.02783557145848216,
        "min_triangle_area=0.0278356, best known = 0.0278."
    )

    test = {
        "solution" : soln.copy(),
        "output" : [[0.63557142, 0.2780757 ],
                    [0.41179303, 0.6500378 ],
                    [0.71410244, 0.29145972],
                    [0.35526134, 0.29541504],
                    [0.38375729, 0.3046403 ],
                    [0.55842241, 0.7181737 ],
                    [0.53782542, 0.32973947],
                    [0.67912473, 0.6429485 ],
                    [0.59160907, 0.55614398],
                    [0.27554739, 0.71103061],
                    [0.48850259, 0.26512655],
                    [0.79653191, 0.71262127],
                    [0.51986098, 0.77731033],
                    [0.75680552, 0.64723116]
                    ]
    }
    logger = ExperimentLogger("test_logger_ignores_no_soln")
    logger.log_solution(test["solution"], test["output"])
    
    last_element = {}
    time.sleep(1)

    with jsonlines.open(os.path.join(logger.dirname, "solutions.jsonl"), "r") as reader:
        for line in reader:
            last_element = line
    
    assert last_element == {}

def test_ignores_solution_without_output():
    solution = soln.copy()
    solution.set_scores(
        0.02783557145848216,
        "min_triangle_area=0.0278356, best known = 0.0278."
    )
    test = {
        "solution" : solution,
        "output" : None
    }
    logger = ExperimentLogger("test_logger_ignores_soln_without_output")
    logger.log_solution(test["solution"], test["output"])
    
    last_element = {}
    time.sleep(1)

    with jsonlines.open(os.path.join(logger.dirname, "solutions.jsonl"), "r") as reader:
        for line in reader:
            last_element = line
    
    assert last_element == {}


import json
import uuid

import numpy as np
import pytest

from llamea import Solution


def test_individual_initialization():
    # Test initialization of an individual
    individual = Solution(
        code="def my_solution(): pass",
        name="MySolution",
        description="This is a test solution.",
        generation=1,
    )

    assert isinstance(individual.id, str)  # Check if ID is a string
    assert uuid.UUID(individual.id)  # Ensure the ID is a valid UUID
    assert individual.code == "def my_solution(): pass"
    assert individual.name == "MySolution"
    assert individual.description == "This is a test solution."
    assert individual.generation == 1
    assert individual.fitness == -np.inf  # Default should be -Inf
    assert individual.feedback == ""  # Default should be empty string
    assert individual.error == ""  # Default should be empty string
    assert individual.task_prompt == ""  # Default task prompt


def test_add_metadata():
    # Test adding metadata to an individual
    individual = Solution()
    individual.add_metadata("source", "LLM-generated")
    individual.add_metadata("version", 1.0)

    assert "source" in individual.metadata
    assert "version" in individual.metadata
    assert individual.metadata["source"] == "LLM-generated"
    assert individual.metadata["version"] == 1.0


def test_copy_individual():
    # Test copying an individual
    individual = Solution(
        code="def my_solution(): pass",
        name="MySolution",
        description="This is a test solution.",
        generation=1,
    )
    individual.add_metadata("key", "value")

    new_individual = individual.copy()

    # Check if new_individual is a copy and has a new unique ID
    assert new_individual.id != individual.id
    assert new_individual.parent_ids == [
        individual.id
    ]  # The parent_id of the new individual should be the original individual's ID
    assert new_individual.code == individual.code
    assert new_individual.name == individual.name
    assert new_individual.description == individual.description
    assert new_individual.metadata == individual.metadata  # Metadata should be copied
    assert new_individual.task_prompt == individual.task_prompt


def test_to_dict():
    # Test converting an individual to a dictionary
    individual = Solution(
        code="def my_solution(): pass",
        name="MySolution",
        description="This is a test solution.",
        generation=1,
    )
    individual.add_metadata("source", "LLM-generated")

    individual_dict = individual.to_dict()

    assert isinstance(individual_dict, dict)
    assert individual_dict["code"] == "def my_solution(): pass"
    assert individual_dict["name"] == "MySolution"
    assert individual_dict["description"] == "This is a test solution."
    assert individual_dict["generation"] == 1
    assert "metadata" in individual_dict
    assert individual_dict["metadata"] == {"source": "LLM-generated"}
    assert individual_dict["task_prompt"] == ""


def test_to_json():
    # Test converting an individual to JSON
    individual = Solution(
        code="def my_solution(): pass",
        name="MySolution",
        description="This is a test solution.",
        generation=1,
    )
    individual.add_metadata("source", "LLM-generated")

    individual_json = individual.to_json()

    assert isinstance(individual_json, str)

    # Convert back to dict to validate content
    individual_dict = json.loads(individual_json)

    assert individual_dict["code"] == "def my_solution(): pass"
    assert individual_dict["name"] == "MySolution"
    assert individual_dict["description"] == "This is a test solution."
    assert individual_dict["generation"] == 1
    assert individual_dict["metadata"] == {"source": "LLM-generated"}
    assert individual_dict["task_prompt"] == ""


def test_mutation():
    # Test mutation prompt assignment
    individual = Solution()
    mutation_prompt = "Refine the strategy of the solution."
    individual.set_operator(mutation_prompt)

    assert individual.operator == mutation_prompt


def test_default_values():
    # Test default values of an individual
    individual = Solution()

    assert individual.code == ""
    assert individual.name == ""
    assert individual.description == ""
    assert individual.configspace is None
    assert individual.fitness == -np.inf
    assert individual.feedback == ""
    assert individual.error == ""
    assert individual.parent_ids == []
    assert individual.metadata == {}
    assert individual.task_prompt == ""



import pytest
import os
import jsonlines
import time
from llamea.utils import prepare_namespace, clean_local_namespace, _add_builtins_into
from llamea.loggers import ExperimentLogger

def test_prepare_namespace_imports_all():
    # Test 1: All libraries are available.
    test1 = {
        "code": """import numpy as np
import random
import math""",
        "allowed" : ["numpy>=0.1.0"],
        "namespace_keys": ["np", "random", "math"]
    }

    soln, potential_issue = prepare_namespace(test1["code"], test1["allowed"])
    assert potential_issue is None
    expected_allowed_list = test1["namespace_keys"]
    _add_builtins_into(expected_allowed_list)
    for key in soln:
        assert key in expected_allowed_list
    
def test_prepare_namespace_returns_generic_objects_for_not_allowed_libraries():
    #Test 2 If a library to be imported is not in the allowed list, test return of generic object.
    test2 = {
        "code": """import numpy as np
import random
import scipy
import sklearn
import math""",
        "allowed" : ["numpy>=0.1.0"],
        "namespace_keys": ["np", "random", "math", "scipy", "sklearn"],
        "potential_issue" : "scipy, sklearn are currently not allowed to be imported in this framework."
    }
    name_space, potential_issue = prepare_namespace(test2["code"], test2["allowed"])

    for key, value in name_space.items():
        assert isinstance(value, object)
        if key in test2["allowed"]:
            assert isinstance(value, type.ModuleType)
        assert key in test2["namespace_keys"]
    assert potential_issue == test2["potential_issue"]


def test_clean_local_namespace_generates_purely_local_namespace():
    global_ns = {
        "imported_library_a" : object,
        "imported_library_b" : object,
        "imported_library_c" : object
    }

    # Simulating exec returns local_ns += global_ns
    local_ns = {
        "imported_library_a" : object,
        "imported_library_b" : object,
        "imported_library_c" : object,
        "instantiated_object": object,
        "local_list" : [1, 2, 34, 101]
    }

    local_ns2 = clean_local_namespace(local_ns, global_ns)
    for key in global_ns:
        assert key not in local_ns2
        assert key not in local_ns #Reference semantics.


def test_prepare_namespace_logs_data():
    test1 = {
        "code": """import numpy as np
import random
import scipy
import sklearn
import math""",
        "allowed" : ["numpy>=0.1.0"],
        "namespace_keys": ["np", "random", "math", "scipy", "sklearn"],
        "potential_issue" : "scipy, sklearn are currently not allowed to be imported in this framework.",
        "import_failure" : ["scipy", "sklearn"]
    }

    test2 = {
        "code": """import numpy as np
import random
import scipy
import sklearn
import Qiskit.Aer
import shapely
import math""",
        "allowed" : ["numpy>=0.1.0"],
        "namespace_keys": ["np", "random", "math", "scipy", "sklearn", "Qiskit", "shapely"],
        "potential_issue" : "scipy, sklearn, Qiskit, shapely are currently not allowed to be imported in this framework.",
        "import_failure" : ["scipy", "sklearn", "Qiskit.Aer", "shapely"]
    }

    logger = ExperimentLogger("test-prepare-namespace")


    potential_issues = []

    
    ns, potential_issue = prepare_namespace(test1["code"], test1["allowed"], logger)
    potential_issues.append(potential_issue)
    for key in ns:
        assert key in test1["namespace_keys"]
    ns, potential_issue = prepare_namespace(test2["code"], test2["allowed"], logger)
    potential_issues.append(potential_issue)
    for key in ns:
        assert key in test2["namespace_keys"]
    
    with jsonlines.open(os.path.join(logger.dirname, "import_failures.jsonl"), "r") as reader:
        count = 0
        logged_objects = []
        for obj in reader:
            logged_objects.append(obj)
            count += 1
    assert count == 2
    for rejected_library in logged_objects[0]["import_misses"]:
        assert rejected_library in test1["import_failure"]
    
    for rejected_library in logged_objects[1]["import_misses"]:
        assert rejected_library in test2["import_failure"]


import os

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

# Load the dataset without the "degrees" column

for problem in ["BP", "TSP", "BBO"]:
    data_path = f"ast/graphstats_{problem}.csv"

    # data_path = 'ast/graphstats.csv'
    data = pd.read_csv(data_path)

    # Drop irrelevant columns
    # data = data.drop(columns=["Degrees", "Pagerank", "Betweenness Centrality", "Clustering Coefficients", "Depths"])
    # Drop non-contributing columns
    data = data.drop(
        columns=[
            "Mean Clustering",
            "Max Clustering",
            "Min Depth",
            "Transitivity",
            "Clustering Variance",
            "mean_complexity",
            "total_complexity",
            "mean_token_count",
            "mean_parameter_count",
            "total_parameter_count",
        ]
    )  # "Eigenvector Centrality"

    data.replace([np.inf, -np.inf], np.nan, inplace=True)

    # data["fitness"].fillna(-0.04, inplace=True)
    # data.loc[data["fitness"] < -0.04, 'fitness'] = -0.04

    # data["fitness"] = 1 + (data["fitness"] * 20)

    # data["fitness"] = minmax_scale(data["fitness"])

    data.fillna(0, inplace=True)

    # Separate metadata and features
    metadata_cols = ["fitness", "LLM", "exp_dir", "alg_id", "parent_id"]
    if "code_diff" in data.columns:
        metadata_cols.append("code_diff")
    features = data.drop(columns=metadata_cols)

    print(problem, features.columns, len(features.columns))

    metadata = data[metadata_cols]

    # Normalize feature columns
    scaler = StandardScaler()
    features_scaled = scaler.fit_transform(features)
    features_scaled_df = pd.DataFrame(features_scaled, columns=features.columns)

    # Combine the normalized features with metadata
    data_scaled = pd.concat(
        [metadata.reset_index(drop=True), features_scaled_df], axis=1
    )

    # Create output directory if not exists
    output_dir = f"ast/{problem}_output_per_llm"
    os.makedirs(output_dir, exist_ok=True)

    # Process data for each LLM
    for llm in data_scaled["LLM"].unique():
        llm_data = data_scaled[data_scaled["LLM"] == llm]
        llm_output = []

        # Assign run numbers per exp_dir
        llm_data["run_number"] = llm_data.groupby("exp_dir").ngroup() + 1

        # Prepare the output
        for _, row in llm_data.iterrows():
            run_number = row["run_number"]
            fitness = row["fitness"]
            features_list = ",".join(map(str, row[features.columns].values))
            llm_output.append([run_number, fitness, features_list])

        # Create DataFrame for the output
        llm_output_df = pd.DataFrame(
            llm_output, columns=["run_number", "fitness", "features"]
        )

        # Sort by run_number
        llm_output_df = llm_output_df.sort_values(by="run_number")

        # Save to CSV
        output_file_path = os.path.join(output_dir, f"{llm}_output.csv")
        llm_output_df.to_csv(output_file_path, index=False, header=False, sep="\t")

    print(f"Output files saved in directory: {output_dir}")
